import os
import csv
import re
import numpy as np
import pandas as pd


from tklib.tkutils import terminate
from tklib.tkdatafile import tkDataFile
from tklib.tkcsv import tkCSV
from tklib.tkexcel import tkExcel
import tklib.tkre as tkre


pattern_int   = re.compile(r'^[-+]?[0-9]+$')
pattern_float = re.compile(r'^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$')


class tkVariousData(tkDataFile):
    def __init__(self, path = None, mode = 'r', OpenFile = 1, data_only = True, **args):
#        super(tkFile, self).__init__(path, mode, **args)
        self.fp    = None
        self.path  = None
        self.mode  = None
        self.path  = self.IfYes(path is not None, path, self.path)
        self.mode  = self.IfYes(mode is not None, mode, self.mode)
        self.update(**args)
        
        self.labelarray    = []
        self.datalistarray = []

    def __del__(self):
        self.Close()

    def __str__(self):
        return self.ClassPath()


    def Read_minimum_matrix(self, close_fp = False, force_numeric = True, usage = None):
        if(tkre.Search(r'\.xlsx$', self.path, flag ='i')):
            excel = tkExcel(self.path)
            self.labelarray, self.datalistarray = excel.Read_minimum_matrix(close_fp = close_fp, force_numeric = force_numeric)
        elif(tkre.Search(r'\.csv$', self.path, flag ='i')):
            csv = tkCSV(self.path)
            self.labelarray, self.datalistarray = csv.Read_minimum_matrix(close_fp = close_fp, force_numeric = force_numeric)
        else:
            terminate("Error in tkVariousData.Read_minimum_matrix(): File type [{}] is not supported."
                    .format(self.path), usage = usage)

        return self.labelarray, self.datalistarray

    def read_data(self, infile, xlabel = 0, ylabel = 1, xmin = None, xmax = None, usage = None):
        print("")
        print(f"Read [{infile}]")
        labels, datalist = self.Read_minimum_matrix(close_fp = True, force_numeric = False, usage = usage)
        xlabel, xin = self.FindDataArray(xlabel, flag = 'i')
        ylabel, yin = self.FindDataArray(ylabel, flag = 'i')

        self.labels   = labels
        self.datalist = datalist
        self.xlabel   = xlabel
        self.ylabel   = ylabel

        self.ndata_all = len(xin)
        print("ndata_all=", self.ndata_all)
        self.x = []
        self.y = []
        self.included_index = []
        for i in range(self.ndata_all):
            if xmin is not None and xmin > xin[i]:
                continue
            if xmax is not None and xmax < xin[i]:
                continue
            
            self.x.append(xin[i])
            self.y.append(yin[i])
            self.included_index.append(i)

        self.ndata = len(self.x)
        self.index = range(self.ndata)

        return labels, datalist

    def to_excel(self, outfile, labels, data_list):
        nx = len(data_list)
        max_ndata = 0
        for d in data_list:
            ndata = len(d)
            if max_ndata < ndata:
                max_ndata = ndata
        for i in range(nx):
            ndata = len(data_list[i])
            if ndata < max_ndata:
                if type(data_list[i]) is not list:
                    data_list[i] = data_list[i].tolist()
                data_list = data_list.copy()
                for j in range(max_ndata - ndata):
                    data_list[i].append(None)

        for i in range(nx):
            dl = data_list[i]
            for j in range(len(dl)):
                t = type(dl[j])
                if t is int or t is float:
                    continue

                if t is str:
                    print(f"dl[{j}]={dl[j]} => ", end = '')
                    if pattern_int.match(dl[j]):
                        dl[j] = int(dl[j])
                    else:
                        if pattern_float.match(dl[j]):
                            dl[j] = float(dl[j])
    
                    print(f"{dl[j]}")

        df = pd.DataFrame(np.array(data_list).T, columns = labels)
        df.to_excel(outfile, index = False)

        if not os.path.exists(outfile):
            print("")
            print(f"Error in tkVariousData.to_excel(): Could not write to [{outfile}].")
            return False

        return True

        