import numpy as np
import itertools
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

import physbo


ndata = 11

#score: HVPI, EHVI, TS
score = 'EHVI'
#score = 'TS'


def vlmop2_minus(x):
    n = x.shape[1]
    y1 = 1 - np.exp(-1 * np.sum((x - 1/np.sqrt(n)) ** 2, axis = 1))
    y2 = 1 - np.exp(-1 * np.sum((x + 1/np.sqrt(n)) ** 2, axis = 1))
    return np.c_[-y1, -y2]


class simulator(object):
    def __init__(self, X):
        self.t = vlmop2_minus(X)

    def __call__( self, action):
        return self.t[action]


a = np.linspace(-2, 2 , ndata)
test_X = np.array(list(itertools.product(a, a)))
test_X.shape

simu = simulator(test_X)

fig = plt.figure(figsize = (8, 8))
ax1  = fig.add_subplot(2, 2, 1)
ax2  = fig.add_subplot(2, 2, 2)
ax3  = fig.add_subplot(2, 2, 3)

# https://bourbaki.biz/fit-colorbar-to-a-graph-on-matplotlib/
div1 = make_axes_locatable(ax1)
cax1 = div1.append_axes("right", size="5%", pad=0.1)
div2 = make_axes_locatable(ax2)
cax2 = div2.append_axes("right", size="5%", pad=0.1)

im1 = ax1.imshow(simu.t[:,0].reshape((ndata, ndata)), vmin=-1.0, vmax=0.0, origin="lower", extent=[-2.0, 2.0, -2.0, 2.0])
ax1.set_title("objective 1")
fig.colorbar(im1, cax = cax1)
ax1.plot([1.0/np.sqrt(2.0)], [1.0/np.sqrt(2.0)], '*')

im2 = ax2.imshow(simu.t[:,1].reshape((ndata, ndata)), vmin=-1.0, vmax=0.0, origin="lower", extent=[-2.0, 2.0, -2.0, 2.0])
ax2.set_title("objective 2")
fig.colorbar(im2, cax = cax2)
ax2.plot([-1.0/np.sqrt(2.0)], [-1.0/np.sqrt(2.0)], '*')

plt.tight_layout()
plt.pause(0.1)

policy = physbo.search.discrete_multi.policy(test_X=test_X, num_objectives=2)
policy.set_seed(0)

res = policy.random_search(max_num_probes=50, simulator=simu, disp_pareto_set=True)
res.fx[0:res.num_runs]

res   = policy.bayes_search(max_num_probes=40, simulator=simu, score = score, interval=10)
front, front_num = res.export_pareto_front()
print("front=", front)
print("front_num=", front_num)

def plot_pareto_front(res):
    front, front_num = res.export_pareto_front()
    dominated = [i for i in range(res.num_runs) if i not in front_num]
    points = res.fx[dominated, :]

#    plt.figure(figsize=(7, 7))
#    fig = plt.figure(figsize = (8, 8))
    ax3  = fig.add_subplot(2, 2, 3)
    ax3.scatter(res.fx[dominated,0], res.fx[dominated,1], c = "blue")
    ax3.scatter(front[:, 0], front[:, 1], c = "red")
    ax3.set_title('Pareto front')
    ax3.set_xlabel('Objective 1')
    ax3.set_ylabel('Objective 2')
    ax3.set_xlim([-1.0,0.0])
    ax3.set_ylim([-1.0,0.0])

    plt.tight_layout()
    plt.pause(0.1)
    plt.show()

#パレート解
plot_pareto_front(res)

#劣解領域の体積
v = res.pareto.volume_in_dominance([-1,-1],[0,0])
print("v=", v)
    