import sys
import os
import configparser
import csv
from pprint import pprint
from math import sqrt
import numpy as np
from scipy import optimize
import tkinter


from tklib.tkparams import tkParams
from tklib.tkapplication_gui import tkApplication_GUI
import tklib.tktkinter as tktkinter


usage_str = '''
""
"Test program for tkinter class"
'''[1:-1]

#=============================
# Treat argments
#=============================
def usage(app):
#    app.usage(infile = cparams.infile)
    if app.usage_str is None:
        return

    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

#=============================
# functions
#=============================
def eprint(textbox, *args):
    print(*args, end = '')
    textbox.insert('end', *args)

#=============================
# csvファイルの読み込み
#=============================
def read_csv(path, app, tkvars):
    i = 0
    x = []
    y = []
    with open(path, "r") as f:
        reader = csv.reader(f)

        for row in reader:
            if i == 0:
                header = row
            else:
                xi = float(row[0])
                x.append(xi)
                y.append(float(row[1]))
            i += 1

        app.header = header
        app.x = x
        app.y = y

    textbox = tkvars['outputtext']
    eprint(textbox, "\n")
    eprint(textbox, "CSV data:\n")
    eprint(textbox, "  header: {}\n".format(header))
    eprint(textbox, "  x: {}\n".format(app.x))
    eprint(textbox, "  y: {}\n".format(app.y))
    return header, x, y


def read_data_file(path, app, tkvars):
    print("")
    print("Read data from [{}]".format(path))
    read_csv(path, app, tkvars)

    tkvars['title'].set(path)
    tkvars['xlabel'].set(app.header[0])
    tkvars['ylabel'].set(app.header[1])
    xrange = [min(app.x), max(app.x)]
    tkvars['x0'].set(xrange[0])
    tkvars['x1'].set(xrange[1])

def read_parameter_file(path, app, tkvars):
    print("")
    print(f"Read parameters from [{path}]")
    cparams.read_parameters(path = path, AddSection = 0, IsPrint = False)
    cparams.print_parameters()

def prm_path_button_click(app, tkvars):
    config = app.configparams
    cparams = app.get_params()

    selpath = tkinter.filedialog.askopenfilename(filetypes = app.prm_file_type, initialdir = config.ini_dir)
    tkvars["prmpath"].set(selpath)
    
    read_parameter_file(selpath, app, tkvars)
    if cparams.get("data_file", None) is not None:
        read_data_file(cparams.data_file, app, tkvars)
        tkvars['x0'].set(cparams.x0)
        tkvars['x1'].set(cparams.x1)

def data_path_button_click(app, tkvars):
    config = app.configparams
    cparams = app.get_params()

    selpath = tkinter.filedialog.askopenfilename(filetypes = app.data_file_type, initialdir = config.ini_dir)
    prmpath = os.path.splitext(selpath)[0] + '.prm'
    cparams.set_attr("data_file", selpath)
    cparams.set_attr("parameter_file", prmpath)
    tkvars["datapath"].set(selpath)
    tkvars["prmpath"].set(prmpath)

    read_data_file(selpath, app, tkvars)

def plot(app, tkvars, xc = None, yc = None):
#=============================
# グラフの表示
#=============================
    fontsize = tkvars['font_size'].get()
    
    canvas = app.canvas
    figure = app.figure
    figure.clf()
    ax = figure.add_subplot(111)
    if app.get('x', None) is not None:
        ax.plot(app.x,  app.y,  label = 'raw data', marker = 'o', linestyle = 'None')
    if xc is not None:
        ax.plot(xc, yc, label = 'fitted', linestyle = 'dashed')
    ax.set_title(tkvars['title'].get(), fontsize = fontsize)
    ax.set_xlabel(tkvars['xlabel'].get(), fontsize = fontsize)
    ax.set_ylabel(tkvars['ylabel'].get(), fontsize = fontsize)
    ax.legend(fontsize = fontsize)
    ax.tick_params(labelsize = fontsize)
    figure.tight_layout()
    canvas.draw()

def plot_button_click(app, tkvars):
    plot(app, tkvars)

#=============================
# scipy.optimize()による最小化
#=============================
def lsq_button_click(app, tkvars):
    cparams = app.get_params()

    eprint(tkvars['outputtext'], "\n")
    eprint(tkvars['outputtext'], "polynomial fit by scipy.optimize() start:\n")

    if app.get('x', None) is None:
        plot(app, tkvars)
        return

    x  = app.x
    y  = app.y
    x0 = tkvars['x0'].get()
    x1 = tkvars['x1'].get()
    xf = []
    yf = []
    for i in range(len(x)):
        if x0 <= x[i] <= x1:
            xf.append(x[i])
            yf.append(y[i])

# leastsqの戻り値は、最適化したパラメータのリストと、最適化の結果
    ret = np.polyfit(xf, yf, deg = tkvars['fit_order'].get(), full = True)
    ai = ret[0]
    res = sqrt(ret[1][0] / len(x))
    
    cparams.set_attr("ai", ai.tolist())
    cparams.set_attr("S2", res)

    eprint(tkvars['outputtext'], "  lsq result: ai={}\n".format(ai))
    eprint(tkvars['outputtext'], "              residual={}\n".format(res))

#=============================
# グラフの表示
#=============================
#表示データの作成
    ncal = 100
    xmin = min(x)
    xmax = max(x)
    xstep = (xmax - xmin) / (ncal - 1)
    xc = []
    yc = []
    for i in range(ncal):
        xi = xmin + i * xstep
        yi = np.poly1d(ai)(xi)
        xc.append(xi)
        yc.append(yi)

    plot(app, tkvars, xc = xc, yc = yc)

#=============================
# File=>openメニュー、pathボタン
#=============================
def inipath_button_click(obj):
    print("obj=", obj)
    selpath = filedialog.askopenfilename(filetypes = file_type, initialdir = ini_dir)
    obj['inipath'].set(selpath)

def path_button_click(obj):
    selpath = filedialog.askopenfilename(filetypes = file_type, initialdir = ini_dir)
    obj['datapath'].set(selpath)
    header, x, y = read_csv(selpath, obj)
    obj['x0'].set(min(x))
    obj['x1'].set(max(x))

def create_menu(app, config, cparams, root_window):
    ctkvars = app.tkvars
    tkvars  = ctkvars.get_param_dict()

    menu_bar  = app.create_menu()

    menu_file = tkinter.Menu(menu_bar, tearoff = 0)
    menu_file.add_command(label = 'Open', accelerator = 'Ctrl+O', 
                command = lambda: data_path_button_click(app, tkvars))
    menu_file.add_command(label='exit', accelerator='Alt+E',  command = lambda: app.on_closing())
    menu_bar.add_cascade(label = 'File', menu = menu_file)

    menu_tool = tkinter.Menu(menu_bar, tearoff = 0)
    menu_tool.add_command(label = 'Setup', accelerator = 'Ctrl+u', 
                command = app.dialog_setup)
    menu_bar.add_cascade(label = 'Tool', menu = menu_tool)

    return menu_bar

def create_toolbar(app, config, cparams, root_window):
    def toolbar_clickevent(event):
        pass

    tool_bar = app.create_toolbar(bg = 'blue')

    setupButton = tktkinter.Button(app, tool_bar, text = "Setup", 
                    command = lambda: app.dialog_setup(entry_width = 50, is_print = True))
    setupButton.pack(side = tkinter.LEFT, anchor = 'w', padx = 1, pady = 0)

    exitButton = tktkinter.Button(app, tool_bar, text = 'Exit', command = lambda: app.on_closing(), anchor = tkinter.NW)
    exitButton.pack(side = tkinter.LEFT, anchor = 'w', padx = 1, pady = 0)


def create_right_frame1(app, config, cparams, pane, title):
    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)

    app.canvas = tktkinter.Canvas(app, frame)

    return frame

def create_right_frame2(app, config, cparams, pane, title):
    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)

    memo_entry = tktkinter.Editor(app, frame,
                    frame_args      = {'bg': 'white'},
                    head_label_args = {'text': 'Path:'},
                    entry_args      = {'textvariable': ['inipath', tkinter.StringVar()], 'width': None},
#                    entry_args      = {'textvariable': obj['inipath'], 'width': None},
                    button_args     = {'text': 'path'},
#                    button_args     = {'text': 'path', 'command': lambda: inipath_button_click(obj)},
                )
    memo_entry.pack(fill = tkinter.BOTH, expand = True)

    return frame

def create_left_frame4(app, config, cparams,  pane, title):
    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)

    treeview = tkinter.ttk.Treeview(frame, columns=["colA", "colB", "colC"])
    treeview.column("#0", width=100)
    treeview.column("colA", width=80)
    treeview.column("colB", width=80)
    treeview.column("colC", width=80)
    treeview.heading("#0", text="階層列")
    treeview.heading("colA", text="データ列A")
    treeview.heading("colB", text="データ列B")
    treeview.heading("colC", text="データ列C")
    itemAId = treeview.insert("",      tkinter.END, text="itemA", open=True, values=("data1a", "data1b", "data1c"))
    itemBId = treeview.insert(itemAId, tkinter.END, text="itemB", open=True, values=("data2a", "data2b", "data2c"))
    itemCId = treeview.insert(itemBId, tkinter.END, text="itemC", values=("data3a", "data3b", "data3c"))
    itemDId = treeview.insert("",      tkinter.END, text="itemD", values=("data4a", "data4b", "data4c"))
    treeview.pack()        

def create_left_frame3(app, config, cparams,  pane, title):
    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)
    textframe = tktkinter.Frame(app, frame, bg='dim gray')
    textframe.pack()
    text = tkinter.Text(textframe, height = 4, width = 20)
    scrollbar = tkinter.Scrollbar(textframe, orient = tkinter.VERTICAL, command=text.yview)
    text["yscrollcommand"] = scrollbar.set
    text.grid(row=0, column=0)
    scrollbar.grid(row=0, column=1, sticky=(tkinter.N, tkinter.S))

def create_left_frame2(app, config, cparams,  pane, title):
    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)

    mb = tkinter.Menubutton(frame, text="menubutton:Select alphabet")
    mb.pack()
    menu = tkinter.Menu(mb)
    menu.add_command(label="A")
    menu.add_command(label="B")
    mb["menu"] = menu

    scale = tkinter.Scale(frame, width = 20, length = 300, label = 'title', 
                orient = tkinter.HORIZONTAL, showvalue = True,
                from_ = 0, to = 100, tickinterval = 20)
    scale.pack()

    class CustomDialog(tkinter.simpledialog.Dialog):
        def __init__(self, master, title=None) -> None:
            super(CustomDialog, self).__init__(parent=master, title=title)

        def body(self, master) -> None:
            entry = tkinter.Entry(master)
            entry.grid(column=0, row=0)
    
        def buttonbox(self):
            box = tkinter.Frame(self)
            self.button1 = tkinter.Button(box, text="OK", width=10, command=self.ok, state=tkinter.DISABLED)
            self.button1.pack(side=tkinter.LEFT, padx=5, pady=5)
            self.button2 = tkinter.Button(box, text="Cancel", width=10, command=self.cancel)
            self.button2.pack(side=tkinter.LEFT, padx=5, pady=5) 
            box.pack() 

        def switchButtonState(self):
            if self.button1['state'] == tk.DISABLED:
                self.button1['state'] = tk.NORMAL 

        def apply(self):  # ダイアログボックスを閉じた後
            pass
#            print(self.var.get()) 

    def display_dialog():
        CustomDialog(pane)
#        tkinter.simpledialog.Dialog(pane)

    button = tktkinter.Button(app, frame, text = "dialog", command = display_dialog)
    button.pack(side = 'top', anchor = 'w', padx = 5, pady = 5)
    
    progressbar = tkinter.ttk.Progressbar(frame, value = 50)
    progressbar.pack()

    check_var = tkinter.BooleanVar(value = True)
#    check_var.set(True)

# イベントハンドラーでget()しないと正常に初期化されない
    def checked_button():
#        print("checked", check_var.get())
        check_var.get()
        
    check = tkinter.Checkbutton(frame, variable = check_var, text = '選択しますか？',
                                onvalue = True, offvalue = False,
                                command = checked_button)
    check.pack()

    var = tkinter.StringVar()
    var.set('aaa')
    optionmenu = tkinter.OptionMenu(frame, var, 'aaa', 'bbb')
    optionmenu.pack()

    combobox = tkinter.ttk.Combobox(frame, values=('aaa', 'bbb', 'ccc'))
    combobox.pack()
        
    rdvar = tkinter.StringVar()
    rdvar.set('B')
    aButton = tkinter.Radiobutton(frame, variable = rdvar, 
                value = 'A', text = 'Option A') #, command = self.showValue)
    aButton.pack()

    bButton = tkinter.Radiobutton(frame, variable = rdvar,
                value = 'B', text = 'Option B') #, command=self.showValue)
    bButton.pack()

    spinbox = tkinter.Spinbox(frame, cursor = 'clock', 
                from_ = 0, to = 100, increment = 3, format = '%12.6f')
    spinbox.pack()

    listbox = tkinter.Listbox(frame)
    for month in ("1月", "2月", "3月", "4月", "5月", "6月", "7月", "8月", "9月", "10月", "11月", "12月"):
        listbox.insert(tkinter.END, month)
    listbox.pack()

    return frame

def create_left_frame1(app, config, cparams,  pane, title):
    ctkvars = app.tkvars
    tkvars  = ctkvars.get_param_dict()

    frame = tktkinter.Frame(app, pane, bg='dim gray')
    pane.add(frame, text = title)

    ltabframe1 = frame

# Parameter path frame
    applications = app.get_external_apps()
    prm_path_frame = tktkinter.PathFrame(app, ltabframe1,
        frame_args      = {'bg': 'white'},
        head_label_args = {'text': 'Parameter file:', 'anchor': tkinter.E},
        entry_args      = {'textvariable': ['prmpath', tkinter.StringVar(value = cparams.get("parameter_file", ""))], 
                           'width': None},
        combobox_args   = {'textvariable': ['prmpath', tkinter.StringVar(value = '')], 
                           'values': ('abc', 'def'),
                           'width': None},
        tail_label_args = {'text': ' '},
        button_args       = {'text': 'path', "width": 4, 'font': ("", 8), 
                            'command': lambda: prm_path_button_click(app, tkvars)},
        edit_button_args  = {'text': 'ed',  "width": 2, 'font': ("", 8)},
        optionmenu_args   = {'textvariable': ['app_path',  tkinter.StringVar(value = 'app')],
                             "width": 3, 'font': ("", 8),
                             'values': applications},
#        menubottom_args   = {'text': 'cmd', "width": 3, 'font': ("", 8)},
        shell_button_args = {'text': 'sh',  "width": 2, 'font': ("", 8)},
        igridrow = 0,
        is_print = True
        )
    prm_path_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

# Data path frame
    data_path_frame = tktkinter.PathFrame(app, ltabframe1,
        frame_args       = {'bg': 'white'},
        head_label_args  = {'text': 'Data file:', 'anchor': tkinter.E},
        entry_args       = {'textvariable': ['datapath', tkinter.StringVar(value = cparams.get("data_file", ""))], 
                            'width': None},
#        tail_label_args = {'text': ' '},
        button_args      = {'text': 'path', "width": 4, 'font': ("", 8),
                            'command': lambda: data_path_button_click(app, tkvars)},
#        edit_button_args  = {'text': 'ed', "width": 2, 'font': ("", 8)},
        optionmenu_args   = {'textvariable': ['app_path',  tkinter.StringVar(value = 'app')],
                             "width": 3, 'font': ("", 8),
                             'values': applications},
#        shell_button_args = {'text': 'sh', "width": 2, 'font': ("", 8)},
        gridframe = prm_path_frame,
        igridrow = 1
        )

# Range frame
    range_frame = tktkinter.RangeFrame(app, ltabframe1,
        frame_args        = {'bg': 'white'},
        head_label_args   = {"text": 'x range:'},
        x0_entry_args     = {"textvariable": ['x0', tkinter.DoubleVar(value = cparams.xrange[0])], 
                             "from_": 0, "to": 100, "increment": 2, "format": '%12.6f',
                             "width": 10},
        center_label_args = {"text": '-'},
        x1_entry_args     = {"textvariable": ['x1', tkinter.DoubleVar(value = cparams.xrange[1])], 
                             "from_": 0, "to": 100, "increment": 2, "format": '%12.6f',
                             "width": 10},
        igridrow = 0
        )
    range_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)
    
    """
    range_frame2 = tktkinter.RangeFrame(app, ltabframe1,
        frame_args        = {'bg': 'white'},
        head_label_args   = {"text": 'yyy range:'},
        x0_entry_args     = {"textvariable": ['y0', tkinter.DoubleVar(value = 100.0)], "width": 10},
        center_label_args = {"text": '-'},
        x1_entry_args     = {"textvariable": ['y1', tkinter.DoubleVar(value = 3)], "width": 10},
        gridframe = range_frame,
        igridrow = 1
        )
    """

# Title / XY label frame
    title_label_frame = tktkinter.LabelEntryFrame(app, ltabframe1,
        frame_args = {'bg': 'white'},
        label_args = {"text": 'Title:'},
        entry_args = {"textvariable": ['title',tkinter. StringVar()], "width": 30},
#        entry_args = {"textvariable": obj['title'], "width": 30},
        igridrow = 0
        )
    title_label_frame.pack(side = tkinter.TOP, anchor = tkinter.W, padx = 2, pady = 2)

    x_label_frame = tktkinter.LabelEntryFrame(app, ltabframe1,
        frame_args = {'bg': 'white'},
        label_args = {"text": 'x label:'},
        entry_args = {"textvariable": ['xlabel', tkinter.StringVar()], "width": 30},
        gridframe = title_label_frame,
        igridrow = 1
        )
#    x_label_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

    y_label_frame = tktkinter.LabelEntryFrame(app, ltabframe1,
        frame_args = {'bg': 'white'},
        label_args = {"text": 'y label:'},
        entry_args = {"textvariable": ['ylabel', tkinter.StringVar()], "width": 30},
        gridframe = title_label_frame,
        igridrow = 2
        )
#    y_label_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

# Order frame
    order_frame = tktkinter.LabelEntryFrame(app, ltabframe1,
        frame_args = {'bg': 'white'},
        label_args = {"text": 'fit order:'},
        entry_args = {"textvariable": ['fit_order', tkinter.IntVar(value = cparams.fit_order)], "width": 30},
        gridframe = title_label_frame,
        igridrow = 3
        )
#    order_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

# Font size frame
    fontsize_frame = tktkinter.LabelEntryFrame(app, ltabframe1,
        frame_args = {'bg': 'white'},
        label_args = {"text": 'Font size:'},
        entry_args = {"textvariable": ['font_size', tkinter.StringVar(value = config.fontsize)], "width": 30},
        gridframe = title_label_frame,
        igridrow = 4
        )
#    fontsize_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

# Button frame
    button_frame = tktkinter.Frame(app, ltabframe1)
    plot_button = tktkinter.Button(app, button_frame, text = 'plot',
        command = lambda: plot_button_click(app, tkvars))
    plot_button.pack(side = 'left')
    
    lsq_button = tktkinter.Button(app, button_frame, text = 'lsq',
        command = lambda: lsq_button_click(app, tkvars))
    lsq_button.pack(side = 'left')
    button_frame.pack(side = tkinter.TOP, anchor = 'w', padx = 2, pady = 2)

# Output text frame
    output_text = tktkinter.Text(app, ltabframe1,
        width = 60,
        height = 10)
    output_text.pack(side = tkinter.TOP, anchor = 'w', expand = True, padx = 2, pady = 2)
    tkvars['outputtext'] = output_text

    return frame

def create_window(app, config, cparams):
    config  = app.configparams
    ctkvars = app.tkvars
    tkvars  = ctkvars.get_param_dict()

    root_window = app.create_window(title = 'title', 
                    geometry = config.geometry, minsize = config.minsize)
    menu_bar = create_menu(app, config, cparams, root_window)
    tool_bar = create_toolbar(app, config, cparams, root_window)
    
# Paned window
    main_pane, left_pane, right_pane \
            = app.PanedWindow2(main_pane_args = {'sashwidth': 2}, 
                               left_pane_args = {'width': config.left_pane_width})

# Left pane
    ltabframe1 = create_left_frame1(app, config, cparams, left_pane, 'Fit')
    ltabframe2 = create_left_frame2(app, config, cparams, left_pane, 'Test1')
    ltabframe3 = create_left_frame3(app, config, cparams, left_pane, 'Scrolled Text')
    ltabframe4 = create_left_frame4(app, config, cparams, left_pane, 'Treeview')

# Right pane
    rtabframe1 = create_right_frame1(app, config, cparams, right_pane, 'Graph')
    rtabframe2 = create_right_frame2(app, config, cparams, right_pane, 'Edit')

# 起動時引数でデータ／設定ファイル名を受け取る
    argv = sys.argv
    if len(argv) >= 2:
        cparams.path = argv[1]
        cparams.data_file      = os.path.splitext(cparams.path)[0] + '.csv'
        cparams.parameter_file = os.path.splitext(cparams.path)[0] + '.prm'
        tkvars['prmpath'].set(cparams.parameter_file)
        tkvars['datapath'].set(cparams.data_file)
        read_parameter_file(cparams.parameter_file, app, tkvars)
        if cparams.get("data_file", None) is not None:
            read_data_file(cparams.data_file, app, tkvars)
            tkvars['x0'].set(cparams.xrange[0])
            tkvars['x1'].set(cparams.xrange[1])


    app.mainloop()

def initialize():
#================================
# Global variables
#================================
    app          = tkApplication_GUI(usage_str = usage_str)
    argv, narg   = app.get_argv()

    ctkvars = app.tkvars
    tkvars  = ctkvars.get_param_dict()

# Non-configurable configuration
    app.prm_file_type  = [('prm file', '*.prm')]
    app.data_file_type = [('data file', '*.csv')]

# Configurable configuration
    config = tkParams()
    app.configparams = config
    config.debug           = 0
    config.print_level     = 0
    config.confirm_on_exit = True
    config.inifile         = os.path.splitext(app.script_path)[0] + '.ini'
    config.ini_dir         = os.path.abspath(os.path.dirname(__file__))
    config.minsize         = [200, 200]
    config.geometry        = "1000x600"
    config.left_pane_width = 500
    config.figsize         = [8, 4]
    config.figsize_plot    = [16, 10]
    config.fontsize        = 16
    config.legend_fontsize = 8
    
    config.editor_path = ''

    print("")
    print(f"Read ini file from [config.inifile]")
    config.read_parameters(path = config.inifile)
    config.print_parameters(heading = "Configure")

    cparams             = app.get_params()

# フィッティングパラメータ初期値。線形最小二乗の場合は適当
    cparams.fit_order = 2
    cparams.ai0    = [0, 0, 0]
    cparams.xrange = [0, 10]

    return app, config, cparams


def main(app, config, cparams):
    ctkvars = app.tkvars
    tkvars  = ctkvars.get_param_dict()

#    updatevars(app, config)
    
    create_window(app, config, cparams)

    config.geometry = app.root_window_geometry
    print("")
    print(f"Save configuration to [{config.inifile}]")
    config.print_parameters(heading = "Configure")
    config.save_parameters(section = 'Configure', sort_by_keys = True, IsPrint = False)

    cparams.parameter_file  = tkvars["prmpath"].get()  #os.path.splitext(cparams.path)[0] + '.prm'
    if cparams.parameter_file != '':
        print(f"Save parameters to [{cparams.parameter_file}]")
        ret = cparams.save_parameters(cparams.parameter_file, section = 'LSQ', 
                sort_by_keys = True, IsPrint = False, update_commandline = False)
        if not ret:
            app.terminate(f"Error in main(): Cannot write to [{cparams.parameter_file}]", usage = usage)

    app.terminate(usage = usage)


if __name__ == "__main__":
    app, config, cparams = initialize()
    main(app, config, cparams)
