import sys
import csv
import numpy as np
from numpy import sin, cos, tan, pi
#import ast
import pandas as pd
import matplotlib.pyplot as plt



from tklib.tkutils import replace_path, getarg, getintarg, getfloatarg, pint, pfloat
from tklib.tksci.tksci import kB, e
from tklib.tkvariousdata import tkVariousData
from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams
from tklib.tksci.tksci import asin, acos, atan, degcos, degsin, degtan, degacos, degasin, degatan, arcsin, arccos, arctan
from tklib.tksci.tksci import factorial, log10, gamma, combination, eVTonm, nmToeV, Bn
from tklib.tksci.tksci import h, h_bar, hbar, e, kB, NA, c, pi, pi2, torad, todeg, basee
from tklib.tksci.tksci import me, mp, mn
from tklib.tksci.tksci import u0, e0, a0, R, F, g, G
from tklib.tksci.tksci import HartreeToeV, RyToeV, KToeV, eVToK, JToeV, eVToJ, Debye


# mode:  'file': read data from csvfile, other: make from func(x)
mode = 'file'

# for mode != 'file'
x0 = 0.0
x1 = 1.0
ndata = 1024

# low and high index limits to remove from FT data
#kcut0 = 5
#kcut1 = 1020
kcut0 =  0
kcut1 = 10

infile  = 'dos.xls'

xlabel = 0
ylabel = 1

app = tkApplication()
cparams = tkParams()
cparams.figsize = [12, 8]
cparams.fontsize = 12
cparams.legend_fontsize = 10

logfile = app.replace_path(infile)
outfile = app.replace_path(infile, template = ["{dirname}", "{filebody}-smoothing-fft.xlsx"])
print(f"Open logfile [{logfile}]")
app.redict(targets = ["stdout", logfile], mode = 'w')


def getarg(iarg, defval = None):
    try:
        return argv[iarg]
    except:
        return defval

def getintarg(iarg, defval = None):
    try:
        return int(argv[iarg])
    except:
        return defval
        
def getfloatarg(iarg, defval = None):
    try:
        return float(argv[iarg])
    except:
        return defval


argv = sys.argv
narg = len(argv)
if narg <= 2:
    print("")
    print("Usage: python smmothing-fft.py csvfile kcut0 kcut1")
    print("")
    exit()
else:
    infile = getarg(1,  infile)
    kcut0  = getfloatarg(2, kcut0)
    kcut1  = getfloatarg(3, kcut1)


# three trigonometric functions with randomized frequencies
# A: amplitude, f: frequency, p1: phase    
def func(x):
    x += 0.03 * np.random.rand();
    f1, p1, A1 = ( 1.5, pi/4.0, 1.0);
    f2, p2, A2 = ( 3.0, pi/3.0, 0.3);
    f3, p3, A3 = (10.0, pi/6.0, 0.5);
    return A1 * sin(2.0*pi * f1 * x + p1) \
         + A2 * sin(2.0*pi * f2 * x + p2) \
         + A3 * sin(2.0*pi * f3 * x + p3);

def main():
    global x0, x1, ndata
    global csvfile
    global outfile

    print("")
    print("========================================")
    print("mode:", mode)
    print("in file:", infile)
    print("k range: {} - {}".format(kcut0, kcut1))
    
    if mode == 'file':
        x = []
        y = []
        print(f"Read [{infile}]")
        datafile = tkVariousData(infile)
        labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = app.usage)
        label_x, x = labels[0], datalist[0] #datafile.FindDataArray(xlabel, flag = 'i')
        label_y, y = labels[1], datalist[1] #datafile.FindDataArray(ylabel, flag = 'i')

        n = len(x)
        xstep = x[1] - x[0]
        x0 = x[1]
        x1 = x0 + xstep * (n-1)
    else:
        x = np.zeros(n)
        y = np.zeros(n)
        print("Make data by func(x) with random scattering")
        xstep = (x1 - x0) / n
        print("x = ({}, {}, {})".format(x0, x1, xstep))
        print("n={}".format(n))
        for i in range(0, n):
            x[i] = x0 + i * xstep
            y[i] = func(x[i])
        n = len(x)

    print("")
    print("ndata:", n)
# FFT
    F = np.fft.fft(y)
    
# calculate FFTed x axis
    xfrange = 1.0 / xstep
    dxf = xfrange / (n-1)
    xf = np.zeros(n)
    for i in range(0, n):
        xf[i] = i * dxf

    print("")
    print("x range     : {} - {}".format(min(x), max(x)))
    print("x(fft) range: {} - {}".format(min(xf), max(xf)))

# cut FT data outside [lf, hf] for smoothing
# note the FFTed data is symmetric for real part and anti-symmetric for imagenary part with respect to n/2
# LF data are near ix = 0 and ix = n - 1 HF data are around n/2
    Fs = np.zeros(n, dtype = 'complex')
    for i in range(0, int(n/2)):
        if kcut0 <= xf[i] <= kcut1:
            Fs[i] = F[i]
            Fs[n - i - 1] = F[n - i - 1]
        else:
#            print("cut: {} {}".format(i, xf[i]))
            Fs[i] = 0.0
            Fs[n - i - 1] = 0.0

# iFFT
    fs = np.fft.ifft(Fs)

    Fr    = [F[i].real  for i in range(n)]
    Fi    = [F[i].imag  for i in range(n)]
    Fsr   = [Fs[i].real for i in range(n)]
    Fsi   = [Fs[i].imag for i in range(n)]
    fsr   = [fs[i].real for i in range(n)]
    fsi   = [fs[i].imag for i in range(n)]
    fsabs = [abs(fs[i]) for i in range(n)]
    df = pd.DataFrame(np.array([x, y, xf, Fr, Fi, Fsr, Fsi, fsr, fsi, fsabs]).T,
                       columns = ['x', 'y', 'x(fft)', 'y(fft).r', 'y(fft).i', 
                                  'ys(fft,cut).r', 'ys(fft,cut).i',
                                  'ys(ifft).r', 'ys(ifft).i', '|ys(ifft)|']
                       )
    print("")
    print(f"Save to [{outfile}]")
    df.to_excel(outfile, index = False) 


    print("")
    print("Plot")
    fig, axes = plt.subplots(2, 2, figsize = cparams.figsize)
    axes = axes.flatten()

#    ax2 = axes[1].twinx()
    axes[0].tick_params(labelsize = cparams.fontsize)
    axes[1].tick_params(labelsize = cparams.fontsize)
    axes[2].tick_params(labelsize = cparams.fontsize)
    axes[3].tick_params(labelsize = cparams.fontsize)

    axes[0].plot(x, y,     label = 'input data', color = 'black', linewidth = 0.5)
    axes[0].plot(x, fsr,   label = 'FT+IFT real', color = 'red', linewidth = 0.5)
    axes[0].set_xlabel(label_x, fontsize = cparams.fontsize)
    axes[0].set_ylabel(label_y, fontsize = cparams.fontsize)
    axes[0].legend(fontsize = cparams.legend_fontsize)

    ins1 = axes[1].plot(xf, Fr, label = 'FT real',      color = 'red', linewidth = 0.5)
    ins2 = axes[1].plot(xf, Fi, label = 'FT imag',      color = 'blue', linewidth = 0.5)
    axes[1].set_xlabel('k', fontsize = cparams.fontsize)
    axes[1].set_ylabel('$FT^*$ (real)', fontsize = cparams.fontsize)
#    ax2.set_ylabel    ('$FT^*$ (imag)', fontsize = cparams.fontsize)
#    ins = ins1 + ins2 + ins3 + ins4
#    axes[1].legend(ins, [l.get_label() for l in ins], fontsize = cparams.legend_fontsize, loc = 'upper center') #loc = 'best')
    axes[1].legend(fontsize = cparams.legend_fontsize)

    axes[2].plot(x, fsr,   label = 'FT+IFT real', color = 'red', linewidth = 0.5)
    axes[2].plot(x, fsi,   label = 'FT+IFT imag', color = 'blue', linewidth = 0.5)
    axes[2].plot(x, fsabs, label = '|FT+IFT|',    color = 'pink', linewidth = 0.5)
    axes[2].set_xlabel(label_x, fontsize = cparams.fontsize)
    axes[2].set_ylabel(label_y, fontsize = cparams.fontsize)
    axes[2].legend(fontsize = cparams.legend_fontsize)

    ins3 = axes[3].plot(xf, Fsr, label = 'FT real(cut)', color = 'red', linewidth = 0.5)
    ins4 = axes[3].plot(xf, Fsi, label = 'FT imag(cut)', color = 'blue', linewidth = 0.5)
    axes[3].set_xlabel('k', fontsize = cparams.fontsize)
    axes[3].set_ylabel('$FT^*$ (real)', fontsize = cparams.fontsize)
    axes[3].legend(fontsize = cparams.legend_fontsize)

    plt.tight_layout()
    plt.pause(0.001)


    app.terminate("\n", pause = True)


if __name__ == "__main__":
    main()
